Unit PRG;

Interface

Uses
 Except,A64SRC,Strings,MOS6510;

Const
 Version='A64 (PRG) v1.0, (c)1998 Paul TOTH <tothpaul@multimania.com>';

 MaxRef=10;

Type
 PReference=^TReference;
 TReference=record
  Name:string[MaxRef];
  Index:word;
  Size:byte;
  Next:PReference;
 end;

 PLabel=^TLabel;
 TLabel=record
  Name:string[MaxRef];
  Index:word;
  Next:PLabel;
 end;

 TPRG=Object(TA64SRC)
  Code:pointer;
  CodeSeg:word;
  CodeSize:word;
  CodeStart:word;
  CodeIndex:word;
  Reference:PReference;
  Labels:PLabel;
  Constructor Create;
  Destructor Destroy;
  Procedure Compile;
  Procedure BasicToken;
  Procedure DataByte;
  Procedure DataWord;
  Procedure LoadBasic;
  Procedure LoadCode;
  Procedure LoadData;
  Procedure NewReference(AReference:string; ASize:byte);
  Procedure FixupReferences;
  Procedure NewLabel(ALabel:string);
  Function  FindLabel(ALabel:string):word;
  Procedure SaveToFile(AFileName:string);
 end;

Implementation

const
 BasicTokens:array[#128..#203] of pchar=(
  'END','FOR','NEXT','DATA','INPUT#','INPUT','DIM','READ','LET','GOTO',
  'RUN','IF','RESTORE','GOSUB','RETURN','REM','STOP','ON','WAIT','LOAD',
  'SAVE','VERIFY','DEF','POKE','PRINT#','PRINT','CONT','LIST','CLR','CMD',
  'SYS','OPEN','CLOSE','GET','NEW',

  'TAB(','TO','FN','SPC(','THEN','NOT','STEP','+','-','*','/','^','AND',
  'OR','>','=','<','SGN','INT','ABS','USR','FRE','POS','SQR','RND','LOG',
  'EXP','COS','SIN','TAN','ATN','PEEK','LEN','STR$','VAL','ASC','CHR$',
  'LEFT$','RIGHT$','MID$',

  'GO'
 );

Constructor TPRG.Create;
 begin
  inherited Create;
  CodeSize:=$FFFF;
  GetMem(Code,CodeSize);
  CodeSeg:=Seg(Code^);
  CodeStart:=$0801;
  CodeIndex:=CodeStart;
  Reference:=nil;
  Labels:=nil;
 end;

Destructor TPRG.Destroy;
 begin
  FreeMem(Code,CodeSize);
  inherited Destroy;
 end;


Procedure TPRG.BasicToken;
 var
  x:byte;
  tk:char;
  tok:string;
 begin
 { "TEXT" }
  if Line[1]='"' then begin
   x:=2;
   while Line[x]<>'"' do inc(x);
   Move(Line[1],mem[CodeSeg:CodeIndex],x);
   inc(CodeIndex,x);
   Delete(Line,1,x);
   exit;
  end;

 { BASIC TOKENS }
  for tk:=#128 to #203 do begin
   tok:=strpas(BasicTokens[tk]);
   if copy(Line,1,length(tok))=tok then begin
    Mem[CodeSeg:CodeIndex]:=ord(tk); inc(CodeIndex);
    delete(Line,1,length(tok));
    if tok='REM' then begin
     x:=length(Line);
     move(Line[1],mem[CodeSeg:CodeIndex],x);
     inc(CodeIndex,x);
     Line:='';
     exit;
    end;
    exit;
   end;
  end;
 { MEMORY REFERENCE }
  if Line[1]='@' then begin
   x:=2;
   while (x<=length(Line))and(Line[x] in ['A'..'Z','_']) do inc(x);
   NewReference(copy(Line,2,x-2),5);
   delete(Line,1,x-1);
   exit;
  end;
 { JUST A CHAR ... }
  Mem[CodeSeg:CodeIndex]:=ord(Line[1]); inc(CodeIndex);
  delete(Line,1,1);
 end;

Procedure TPRG.LoadBasic;
 var
  lastline:word;
  x:byte;
 begin
  if CodeIndex<>$0801 then raise('BASIC should start at $0801');

  LastLine:=0;

  While GetBasicLine do begin

  { next line }
   if LastLine<>0 then MemW[CodeSeg:LastLine]:=CodeIndex;
   LastLine:=CodeIndex;
   MemW[CodeSeg:LastLine]:=0; inc(CodeIndex,2);

  { set line number }
   MemW[CodeSeg:CodeIndex]:=LineNumber; inc(CodeIndex,2);

  { BASIC }
   while Line<>'' do BasicToken;
   Mem[CodeSeg:CodeIndex]:=0; inc(CodeIndex);

  end;

  if LastLine=0 then raise('Invalide BASIC program');

 { END OF BASIC }
  MemW[CodeSeg:LastLine]:=CodeIndex;
  MemW[CodeSeg:CodeIndex]:=0;
  Inc(CodeIndex,2);
 end;

Procedure TPRG.LoadCode;
 var
  ACode:TCode;
 begin
  NextToken;

  while (Token<>'.')and(Token<>'END') do begin

   if Token[1]='@' then begin

    Token:=''; ReadAlpha;
    if (Token='')or(DropChar<>':') then raise('Invalide label');
    NewLabel(Token);

   end else begin

    NextCodeLine;

    Encode(Token,Line,ACode);
    Mem[CodeSeg:CodeIndex]:=ACode.Code; inc(CodeIndex);

    if line[1]='#' then begin
     case line[2] of
      '<': begin
            Delete(line,1,2);
            NewReference(line,3);
           end;
      '>': begin
            Delete(line,1,2);
            NewReference(line,4);
           end;
     end;
    end;

    if line[1]='@' then begin
     delete(line,1,1);
     case ACode.Size of
      2:NewReference(line,1);
      3:NewReference(line,2);
      else raise('Internal error');
     end;
    end;

    case ACode.Size of
     2 : begin Mem[CodeSeg:CodeIndex]:=ACode.Pip1; inc(CodeIndex); end;
     3 : begin MemW[CodeSeg:CodeIndex]:=ACode.Pip2; inc(CodeIndex,2); end;
    end;

   end;
   NextToken;
  end;

 end;

Procedure TPRG.DataByte;
 var
  c:char;
  value:longint;
  s:string;
 begin
  SkipSpaces;
  c:=NextChar;
  While c<>#13 do begin
   if c=';' then begin NextLine; exit end;
   if c in ['''','"'] then begin
    s:=LoadString;
    Move(s[1],Mem[CodeSeg:CodeIndex],length(s));
    inc(CodeIndex,length(s));
   end else begin
    if not GetValue(value) then raise('Invalide Data');
    if (Value>255)or(Value<-128) then raise('Data out of range');
    Mem[CodeSeg:CodeIndex]:=byte(value); inc(CodeIndex);
   end;
   SkipSpaces;
   c:=NextChar;
   if not (c in [#13,',',';']) then raise('"," expected');
   if c=',' then begin
    DropChar;
    SkipBlanks;
    c:=NextChar;
   end;
  end;
 end;

Procedure TPRG.DataWord;
 var
  value:longint;
 begin
  While GetValue(value) do begin
   if (Value>65535)or(Value<-32768) then raise('Invalide data');
   MemW[CodeSeg:CodeIndex]:=word(value); inc(CodeIndex,2);
   SkipSpaces;
   if NextChar<>',' then break;
   DropChar;
  end;
 end;

Procedure TPRG.LoadData;
 begin
  NextToken;
  While (Token<>'.')and(Token<>'END') do begin
   if Token='DB' then DataByte else
   if Token='DW' then DataWord else
   NewLabel(Token);
   NextToken;
  end;
 end;

Procedure TPRG.NewReference(AReference:string;ASize:byte);
 var
  ref:PReference;
 begin
  if length(AReference)>MaxRef then raise('name too long');
  new(ref);
  ref^.Name:=AReference;
  ref^.Index:=CodeIndex;
  ref^.Next:=Reference;
  ref^.Size:=ASize;
  Reference:=Ref;
  if ASize=5 then begin
   FillChar(Mem[CodeSeg:CodeIndex],5,'0');
   inc(CodeIndex,5);
  end;
 end;

Procedure TPRG.FixupReferences;
 var
  ref:PReference;
  adr:word;
  d:word;
  x:byte;
 begin
  ref:=Reference;
  while ref<>nil do begin
   adr:=FindLabel(Ref^.Name);
   case ref^.Size of
    1:Mem[CodeSeg:Ref^.Index]:=byte(adr-Ref^.Index-1);
    2:MemW[CodeSeg:Ref^.Index]:=adr;
    3:Mem[CodeSeg:Ref^.Index]:=adr and 255;
    4:Mem[CodeSeg:Ref^.Index]:=adr shr 8;
    5:begin
       d:=10000;
       for x:=0 to 4 do begin
        inc(Mem[CodeSeg:Ref^.Index+x],(adr div d) mod 10);
        d:=d div 10;
       end;
      end;
   end;
   ref:=ref^.Next;
  end;
 end;

Procedure TPRG.NewLabel(ALabel:string);
 var
  lbl:PLabel;
 begin
  if length(ALabel)>MaxRef then raise('label name too long');
  lbl:=Labels;
  while lbl<>nil do begin
   if lbl^.Name=ALabel then raise('Duplicate label');
   lbl:=lbl^.Next;
  end;
  new(lbl);
  lbl^.Name:=ALabel;
  lbl^.Index:=CodeIndex;
  lbl^.Next:=Labels;
  Labels:=lbl;
 end;

Function TPRG.FindLabel(ALabel:string):word;
 var
  lbl:PLabel;
 begin
  lbl:=Labels;
  while lbl<>nil do begin
   if lbl^.Name=ALabel then begin
    FindLabel:=lbl^.Index;
    exit;
   end;
   lbl:=lbl^.Next;
  end;
  writeln(ALabel);
  raise('Label not found');
 end;

Procedure TPRG.Compile;
 begin
  NextToken;
  repeat
   if Token<>'.' then raise('"." expected');
   NextToken; NextLine;
   if Token='BASIC'then LoadBasic else
   if Token='CODE' then LoadCode else
   if Token='DATA' then LoadData else raise('Unknow identifier');
  Until Token='END';
  FixupReferences;
 end;

Procedure TPRG.SaveToFile(AFileName:string);
 var
  f:file;
 begin
  assign(f,AFileName);
  rewrite(f,1);
  blockwrite(f,CodeStart,2);
  blockwrite(f,Mem[CodeSeg:CodeStart],CodeIndex-CodeStart);
  close(f);
 end;

end.