Unit Source;
{ C64 Assembler (c) 1998, Paul TOTH <tothpaul@multimania.com>

 http://www.multimania.com/tothpaul

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
}

Interface

Uses
 Except;

{------- TSource ----------------------}
Type
 TSource=Object
  Source:PChar;
  SourceSize:word;
  SourceIndex:word;
  SourceLine:word;
  LineStart:word;
  Token:string;
  Constructor Create;
  Destructor Destroy;
  Procedure ErrorLine;
  Procedure LoadFromFile(AFileName:string);
  function NextLine:string;
  procedure SkipBlanks;
  procedure SkipSpaces;
  function DropChar:char;
  function NextChar:char;
  procedure ReadAlpha;
  procedure ReadNumber;
  function GetValue(var Value:longint):boolean;
  function LoadString:string;
 end;

const
 digits:array[0..$F] of char='0123456789ABCDEF';

Implementation

Constructor TSource.Create;
 begin
  SourceSize:=0;
  SourceIndex:=0;
  SourceLine:=1;
  LineStart:=0;
 end;

Destructor TSource.Destroy;
 begin
  if SourceSize>0 then FreeMem(Source,SourceSize);
 end;

Procedure TSource.ErrorLine;
 var
  i:word;
  c:char;
 begin
  if SourceIndex>0 then begin
   writeLn('Error in line ',SourceLine);
   for i:=LineStart to SourceSize-1 do begin
    c:=Source[i];
    if c=#10 then break;
    write(c);
   end;
   writeln;
   for i:=LineStart to SourceIndex-1 do write(' '); writeln('^');
  end;
 end;

Procedure TSource.LoadFromFile(AFileName:string);
 var
  TextFile:file;
  fSize:longint;
 begin
  Assign(TextFile,AFileName);
  Reset(TextFile,1);
  if IoResult<>0 then raise('File not found');
  fSize:=FileSize(TextFile);
  if fSize>65535-8 then begin
   Close(TextFile);
   raise('File too big');
  end;
  SourceSize:=fSize;
  GetMem(Source,SourceSize);
  BlockRead(TextFile,Source^,SourceSize);
  Close(TextFile);
 end;
(*
Function TA64SRC.NextToken:string;
 begin
  repeat
   SkipBlanks;
   Token:=UpCase(DropChar);
   case Token[1] of
    'A'..'Z','_': ReadAlpha;
    '0'..'9'    : ReadNumber;
    '('         : if NextChar='*' then Token:=Token+DropChar;
{    '/'         : if NextChar in ['*','/'] then Token:=Token+DropChar; }
    '>',':'     : if NextChar='=' then Token:=Token+DropChar;
    '<'         : if NextChar='>' then Token:=Token+DropChar else begin
                   if NextChar='=' then begin
                    Token:=Token+DropChar;
                    if NextChar='>' then Token:=Token+DropChar;
                   end;
                  end;
   end;
   if Token='DIV' then Token:='/' else
   if Token='{'  then begin Token:=''; Comment(1) end else { Pascal comments }
   if Token='(*' then begin Token:=''; Comment(2) end else { Pascal comments }
   if Token=';'  then begin Token:=''; Comment(3) end else { C/Delphi comments }
   if Token='/*' then begin Token:=''; Comment(4) end else { C comments }
   NextToken:=Token;
  until Token<>'';
 end;
*)
Function TSource.NextLine:string;
 var
  Result:string;
  l:word;
 begin
  Result:='';
  l:=SourceLine;
  repeat Result:=Result+DropChar until (SourceLine>l);
  dec(Result[0]); {#13}
  NextLine:=Result;
 end;
(*
Procedure TSource.Comment(Style:byte);
 var
  l:word;
 begin
  if (Style<3)and(NextChar='$') then begin
 { Compiler options }
   case Style of
    1 : repeat Until DropChar='}';
    2 : begin
         repeat Until (DropChar='*') and (NextChar=')');
         DropChar;
        end;
   end;
  end else begin
 { Just a comment }
   case Style of
    1 : repeat Until DropChar='}';
    2 : begin
         repeat Until (DropChar='*') and (NextChar=')');
         DropChar;
        end;
    3 : begin
         l:=SourceLine;
         repeat DropChar Until (SourceLine>l);
        end;
    4 : begin
         repeat Until (DropChar='*') and (NextChar='/');
         DropChar;
        end;
   end;
  end;
 end;
*)
Procedure TSource.SkipBlanks;
 var
  c:char;
 begin
  While (SourceIndex<SourceSize) do begin
   c:=Source[SourceIndex];
   if c=#13 then inc(SourceLine);
   if c=#10 then LineStart:=SourceIndex+1;
   if c in [#9,#10,#13,' '] then
    Inc(SourceIndex)
   else
    exit;
  end;
 end;

Procedure TSource.SkipSpaces;
 var
  c:char;
 begin
  While (SourceIndex<SourceSize) do begin
   if  Source[SourceIndex] in [#9,' '] then
    Inc(SourceIndex)
   else
    exit;
  end;
 end;

Function TSource.DropChar:Char;
 var
  c:char;
 begin
  c:=NextChar;
  if c=#0 then raise('Unexpected end of file');
  inc(SourceIndex);
  if c=#13 then inc(SourceLine);
  if c=#10 then LineStart:=SourceIndex+1;
  DropChar:=c;
 end;

Function TSource.NextChar:Char;
 begin
  if SourceIndex<SourceSize then
   NextChar:=Source[SourceIndex] else NextChar:=#0;
 end;

Procedure TSource.ReadAlpha;
 var
  c:char;
 begin
  c:=UpCase(NextChar);
  while c in ['A'..'Z','_'] do begin
   Token:=Token+c; inc(SourceIndex);
   ReadNumber;
   c:=UpCase(NextChar);
  end;
 end;

Procedure TSource.ReadNumber;
 var
  c:char;
 begin
  c:=UpCase(NextChar);
  while c in ['0'..'9'] do begin
   Token:=Token+c; inc(SourceIndex);
   c:=UpCase(NextChar);
  end;
 end;

Function TSource.GetValue(Var Value:longint):boolean;
 var
  base:byte;
  neg:boolean;
  c:char;
  p:byte;
 begin
  SkipSpaces;
  Value:=0;
  neg:=false;
  base:=10;
  if NextChar='-' then begin
   neg:=true;
   inc(SourceIndex);
   SkipSpaces;
  end;
  if NextChar='$' then begin
   base:=16;
   inc(SourceIndex);
   SkipSpaces;
  end;
  c:=NextChar;
  p:=pos(c,Digits);
  GetValue:=False;
  while (p>0)and(p<base) do begin
   GetValue:=True;
   Value:=Base*Value+p-1;
   inc(SourceIndex);
   c:=NextChar;
   p:=pos(c,Digits);
  end;
  if neg then Value:=-Value;
 end;

Function TSource.LoadString:string;
 var
  result:string;
  sep:char;
  c:char;
  done:boolean;
 begin
  sep:=DropChar; { " or ' }
  Result:='';
  done:=false;
  c:=DropChar;
  Repeat
   if c=#13 then raise('String constant exceeds line');
   if (c=sep)and(NextChar<>sep) then
    Done:=True
   else begin
    if length(result)=255 then raise('String too long (255)');
    Result:=Result+c;
    if c=sep then DropChar;
    c:=DropChar;
   end;
  until done;
  LoadString:=Result;
 end;

(*
procedure TSource.EOI;
 begin
  SkipBlanks;
  if DropChar<>';' then raise('";" expected');
  NextToken;
 end;
*)

end.